# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

import attr
from marshmallow3_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class PopularTable:
    """
    DEPRECATED. Use TableSummary
    """
    database: str = attr.ib()
    cluster: str = attr.ib()
    schema: str = attr.ib()
    name: str = attr.ib()
    description: Optional[str] = None


class PopularTableSchema(AttrsSchema):
    """
    DEPRECATED. Use TableSummary
    """
    class Meta:
        target = PopularTable
        register_as_scheme = True



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional, List

from amundsen_common.models.badge import Badge

import attr
from marshmallow3_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class LineageItem:
    key: str  # down/upstream table/col/task key
    level: int  # upstream/downstream distance from current resource
    source: str  # database this resource is from
    badges: Optional[List[Badge]] = None
    usage: Optional[int] = None  # statistic to sort lineage items by
    parent: Optional[str] = None  # key of the parent entity, used to create the relationships in graph


class LineageItemSchema(AttrsSchema):
    class Meta:
        target = LineageItem
        register_as_scheme = True


@attr.s(auto_attribs=True, kw_only=True)
class Lineage:
    key: str  # current table/col/task key
    direction: str  # upstream/downstream/both
    depth: int  # how many levels up/down 0 == all
    upstream_entities: List[LineageItem]  # list of upstream entities
    downstream_entities: List[LineageItem]  # list of downstream entities


class LineageSchema(AttrsSchema):
    class Meta:
        target = Lineage
        register_as_scheme = True



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

import attr

from marshmallow3_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class GenerationCode:
    key: Optional[str]
    text: str
    source: Optional[str]


class GenerationCodeSchema(AttrsSchema):
    class Meta:
        target = GenerationCode
        register_as_scheme = True



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import attr
from http import HTTPStatus
from typing import Any, Dict

from marshmallow3_annotations.ext.attrs import AttrsSchema

OK = 'ok'
FAIL = 'fail'
_ELIGIBLE_HEALTH_CHECKS = [OK, FAIL]
_HEALTH_CHECK_HTTP_STATUS_MAP = {
    OK: HTTPStatus.OK,
    FAIL: HTTPStatus.SERVICE_UNAVAILABLE
}


@attr.s(auto_attribs=True, kw_only=True)
class HealthCheck:
    status: str = attr.ib()
    checks: Dict[str, Any] = dict()

    @status.validator
    def vaildate_status(self, attribute: str, value: Any) -> None:
        if value not in _ELIGIBLE_HEALTH_CHECKS:
            raise ValueError(f"status must be one of {_ELIGIBLE_HEALTH_CHECKS}")

    def get_http_status(self) -> int:
        return _HEALTH_CHECK_HTTP_STATUS_MAP[self.status]

    def dict(self) -> Dict[str, Any]:
        return attr.asdict(self)


class HealthCheckSchema(AttrsSchema):
    class Meta:
        target = HealthCheck
        register_as_scheme = True



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import attr

from marshmallow3_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class Tag:
    tag_type: str
    tag_name: str


class TagSchema(AttrsSchema):
    class Meta:
        target = Tag
        register_as_scheme = True



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional, Dict

import attr

from marshmallow3_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class Filter:
    name: str
    values: List[str]
    operation: str


class FilterSchema(AttrsSchema):
    class Meta:
        target = Filter
        register_as_scheme = True


@attr.s(auto_attribs=True, kw_only=True)
class SearchRequest:
    query_term: str
    resource_types: List[str] = []
    page_index: Optional[int] = 0
    results_per_page: Optional[int] = 10
    filters: List[Filter] = []


class SearchRequestSchema(AttrsSchema):
    class Meta:
        target = SearchRequest
        register_as_scheme = True


@attr.s(auto_attribs=True, kw_only=True)
class SearchResponse:
    msg: str
    page_index: int
    results_per_page: int
    results: Dict  # type: ignore
    status_code: int


class SearchResponseSchema(AttrsSchema):
    class Meta:
        target = SearchResponse
        register_as_scheme = True


@attr.s(auto_attribs=True, kw_only=True)
class UpdateDocumentRequest:
    resource_key: str
    resource_type: str
    field: str
    value: Optional[str]
    operation: str  # can be add or overwrite


class UpdateDocumentRequestSchema(AttrsSchema):
    class Meta:
        target = UpdateDocumentRequest
        register_as_scheme = True



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import textwrap

# Specifying default mapping for elasticsearch index
# Documentation: https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping.html
# Setting type to "text" for all fields that would be used in search
# Using Simple Analyzer to convert all text into search terms
# https://www.elastic.co/guide/en/elasticsearch/reference/current/analysis-simple-analyzer.html
# Standard Analyzer is used for all text fields that don't explicitly specify an analyzer
# https://www.elastic.co/guide/en/elasticsearch/reference/current/analysis-standard-analyzer.html
TABLE_INDEX_MAP = textwrap.dedent(
    """
    {
        "settings": {
            "analysis": {
                "normalizer": {
                    "column_names_normalizer": {
                        "type": "custom",
                        "filter": ["lowercase"]
                    }
                }
            }
        },
        "mappings": {
            "properties": {
                "name": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "schema": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "display_name": {
                    "type": "keyword"
                },
                "last_updated_timestamp": {
                    "type": "date",
                    "format": "epoch_second"
                },
                "description": {
                    "type": "text",
                    "analyzer": "simple"
                },
                "column_names": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword",
                            "normalizer": "column_names_normalizer"
                        }
                    }
                },
                "column_descriptions": {
                    "type": "text",
                    "analyzer": "simple"
                },
                "tags": {
                    "type": "keyword"
                },
                "badges": {
                    "type": "keyword"
                },
                "cluster": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "database": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "key": {
                    "type": "keyword"
                },
                "total_usage": {
                    "type": "long"
                },
                "unique_usage": {
                    "type": "long"
                },
                "programmatic_descriptions": {
                    "type": "text",
                    "analyzer": "simple"
                }
            }
        }
    }
    """
)

DASHBOARD_ELASTICSEARCH_INDEX_MAPPING = textwrap.dedent(
    """
    {
        "settings": {
            "analysis": {
                "normalizer": {
                    "lowercase_normalizer": {
                        "type": "custom",
                        "char_filter": [],
                        "filter": ["lowercase", "asciifolding"]
                    }
                }
            }
        },
        "mappings": {
            "properties": {
                "group_name": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword",
                            "normalizer": "lowercase_normalizer"
                        }
                    }
                },
                "name": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword",
                            "normalizer": "lowercase_normalizer"
                        }
                    }
                },
                "description": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "group_description": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "query_names": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "chart_names": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "tags": {
                    "type": "keyword"
                },
                "badges": {
                    "type": "keyword"
                }
            }
        }
    }
    """
)

USER_INDEX_MAP = textwrap.dedent(
    """
    {
        "mappings": {
            "properties": {
                "email": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "first_name": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "last_name": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "full_name": {
                    "type": "text",
                    "analyzer": "simple",
                    "fields": {
                        "raw": {
                            "type": "keyword"
                        }
                    }
                },
                "total_read": {
                    "type": "long"
                },
                "total_own": {
                    "type": "long"
                },
                "total_follow": {
                    "type": "long"
                }
            }
        }
    }
    """
)

FEATURE_INDEX_MAP = textwrap.dedent(
    """
    {
    "settings": {
        "analysis": {
            "normalizer": {
                "lowercase_normalizer": {
                    "type": "custom",
                    "filter": ["lowercase"]
                }
            }
        }
    },
    "mappings": {
        "properties": {
            "feature_group": {
                "type": "text",
                "analyzer": "simple",
                "fields": {
                    "raw": {
                        "type": "keyword",
                        "normalizer": "lowercase_normalizer"
                    }
                }
            },
            "feature_name": {
                "type": "text",
                "analyzer": "simple",
                "fields": {
                    "raw": {
                        "type": "keyword",
                        "normalizer": "lowercase_normalizer"
                    }
                }
            },
            "version": {
                "type": "keyword",
                "normalizer": "lowercase_normalizer"
            },
            "key": {
                "type": "keyword"
            },
            "total_usage": {
                "type": "long"
            },
            "status": {
                "type": "keyword"
            },
            "entity": {
                "type": "keyword"
            },
            "description": {
                "type": "text"
            },
            "availability": {
                "type": "text",
                "analyzer": "simple",
                "fields": {
                    "raw": {
                        "type": "keyword"
                    }
                }
            },
            "badges": {
                "type": "keyword"
            },
            "tags": {
                "type": "keyword"
            },
            "last_updated_timestamp": {
                "type": "date",
                "format": "epoch_second"
            }
        }
    }
}
"""
)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import attr

from marshmallow3_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class Badge:
    badge_name: str = attr.ib()
    category: str = attr.ib()


class BadgeSchema(AttrsSchema):
    class Meta:
        target = Badge
        register_as_scheme = True



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
import abc
import re
from typing import Any, Dict, Optional, Set


class AtlasStatus:
    ACTIVE = "ACTIVE"
    DELETED = "DELETED"


class AtlasCommonParams:
    qualified_name = 'qualifiedName'
    guid = 'guid'
    attributes = 'attributes'
    relationships = 'relationshipAttributes'
    uri = 'entityUri'
    type_name = 'typeName'
    unique_attributes = 'uniqueAttributes'
    created_timestamp = 'createdTimestamp'
    last_modified_timestamp = 'lastModifiedTimestamp'


class AtlasCommonTypes:
    bookmark = 'Bookmark'
    user = 'User'
    reader = 'Reader'
    cluster = 'Cluster'
    application = 'Application'
    data_set = 'DataSet'

    # These are just `virtual` types which do not actually exist in Atlas.
    # We use those constant values to distinguish Atlas Python Client methods which should be used for populating
    # such data.
    # Tags are published using Glossary API, badges using Classification API. Other entities are published using regular
    # Entity API.
    tag = 'Tag'
    badge = 'Badge'
    resource_report = 'Report'


class AtlasTableTypes:
    table = 'Table'
    column = 'Column'
    database = 'Database'
    schema = 'Schema'
    source = 'Source'
    watermark = 'TablePartition'
    process = 'LineageProcess'


class AtlasDashboardTypes:
    metadata = 'Dashboard'
    group = 'DashboardGroup'
    query = 'DashboardQuery'
    chart = 'DashboardChart'
    execution = 'DashboardExecution'


class AtlasKey(abc.ABC):
    """
    Class for unification of entity keys between Atlas and Amundsen ecosystems.

    Since Atlas can be populated both by tools from 'Atlas world' (like Apache Atlas Hive hook/bridge) and Amundsen
    Databuilder (and each of the approach has a different way to render unique identifiers) we need such class
    to serve as unification layer.
    """

    def __init__(self, raw_id: str, database: Optional[str] = None) -> None:
        self._raw_identifier = raw_id
        self._database = database

    @property
    def is_qualified_name(self) -> bool:
        """
        Property assessing whether raw_id is qualified name.

        :returns: -
        """
        if self.atlas_qualified_name_regex.match(self._raw_identifier):
            return True
        else:
            return False

    @property
    def is_amundsen_key(self) -> bool:
        """
        Property assessing whether raw_id is amundsen key.

        :returns: -
        """
        if self.amundsen_key_regex.match(self._raw_identifier):
            return True
        else:
            return False

    def get_details(self) -> Dict[str, str]:
        """
        Collect as many details from key (either qn or amundsen key)

        :returns: dictionary of entity properties derived from key
        """
        if self.is_qualified_name:
            return self._get_details_from_qualified_name()
        elif self.is_amundsen_key:
            return self._get_details_from_key()
        else:
            raise ValueError(f'Value is neither valid qualified name nor amundsen key: {self._raw_identifier}')

    def _get_details(self, pattern: Any) -> Dict[str, str]:
        """
        Helper function collecting data from regex match

        :returns: dictionary of matched regex groups with their values
        """
        try:
            result = pattern.match(self._raw_identifier).groupdict()

            return result
        except KeyError:
            raise KeyError

    def _get_details_from_qualified_name(self) -> Dict[str, str]:
        """
        Collect as many details from qualified name

        :returns: dictionary of entity properties derived from qualified name
        """
        try:
            return self._get_details(self.atlas_qualified_name_regex)
        except KeyError:
            raise ValueError(f'This is not valid qualified name: {self._raw_identifier}')

    def _get_details_from_key(self) -> Dict[str, str]:
        """
        Collect as many details from amundsen key

        :returns: dictionary of entity properties derived from amundsen key
        """
        try:
            return self._get_details(self.amundsen_key_regex)
        except KeyError:
            raise ValueError(f'This is not valid qualified name: {self._raw_identifier}')

    @property
    @abc.abstractmethod
    def atlas_qualified_name_regex(self) -> Any:
        """
        Regex for validating qualified name (and collecting details from qn parts)

        :returns: -
        """
        pass

    @property
    @abc.abstractmethod
    def amundsen_key_regex(self) -> Any:
        """
        Regex for validating amundsen key (and collecting details from key parts)

        :returns: -
        """
        pass

    @property
    @abc.abstractmethod
    def qualified_name(self) -> str:
        """
        Properly formatted qualified name

        :returns: -
        """
        pass

    @property
    @abc.abstractmethod
    def amundsen_key(self) -> str:
        """
        Properly formetted amundsen key

        :returns: -
        """
        pass

    @property
    def native_atlas_entity_types(self) -> Set[str]:
        """
        Atlas can be populated using two approaches:
        1. Using Atlas-provided bridge/hook tools to ingest data in push manner (like Atlas Hive Hook)
        2. Using Amundsen-provided databuilder framework in pull manner

        Since Atlas-provided tools follow different approach for rendering qualified name than databuilder does,
        to provide compatibility for both approaches we need to act differently depending whether the table entity
        was loaded by Atlas-provided or Amundsen-provided tools. We distinguish them by entity type - in Atlas the
        naming convention assumes '_table' suffix in entity name while Amundsen does not have such suffix.

        If the entity_type (database in Amundsen lingo) is one of the values from this property, we treat it like
        it was provided by Atlas and follow Atlas qualified name convention.

        If the opposite is true - we treat it like it was provided by Amundsen Databuilder, use generic entity types
        and follow Amundsen key name convention.
        """
        return {'hive_table'}

    @property
    def entity_type(self) -> str:
        if self.is_qualified_name:
            return self._database or ''
        else:
            return self.get_details()['database'] \
                if self.get_details()['database'] in self.native_atlas_entity_types else 'Table'


class AtlasTableKey(AtlasKey):
    @property
    def atlas_qualified_name_regex(self) -> Any:
        return re.compile(r'^(?P<schema>.*?)\.(?P<table>.*)@(?P<cluster>.*?)$', re.X)

    @property
    def amundsen_key_regex(self) -> Any:
        return re.compile(r'^(?P<database>.*?)://(?P<cluster>.*)\.(?P<schema>.*?)\/(?P<table>.*?)$', re.X)

    @property
    def qualified_name(self) -> str:
        if not self.is_qualified_name and self.get_details()['database'] in self.native_atlas_entity_types:
            spec = self._get_details_from_key()

            schema = spec['schema']
            table = spec['table']
            cluster = spec['cluster']

            return f'{schema}.{table}@{cluster}'
        else:
            return self._raw_identifier

    @property
    def amundsen_key(self) -> str:
        if self.is_qualified_name:
            spec = self._get_details_from_qualified_name()

            schema = spec['schema']
            table = spec['table']
            cluster = spec['cluster']

            return f'{self._database}://{cluster}.{schema}/{table}'
        elif self.is_amundsen_key:
            return self._raw_identifier
        else:
            raise ValueError(f'Value is neither qualified name nor amundsen key: {self._raw_identifier}')


class AtlasColumnKey(AtlasKey):
    @property
    def atlas_qualified_name_regex(self) -> Any:
        return re.compile(r'^(?P<schema>.*?)\.(?P<table>.*?)\.(?P<column>.*?)@(?P<cluster>.*?)$', re.X)

    @property
    def amundsen_key_regex(self) -> Any:
        return re.compile(r'^(?P<database>.*?)://(?P<cluster>.*)\.(?P<schema>.*?)\/(?P<table>.*?)\/(?P<column>.*)$',
                          re.X)

    @property
    def qualified_name(self) -> str:
        if self.is_amundsen_key:
            spec = self._get_details_from_key()

            schema = spec['schema']
            table = spec['table']
            cluster = spec['cluster']
            column = spec['column']

            return f'{schema}.{table}.{column}@{cluster}'
        elif self.is_qualified_name:
            return self._raw_identifier
        else:
            raise ValueError(f'Value is neither qualified name nor amundsen key: {self._raw_identifier}')

    @property
    def amundsen_key(self) -> str:
        if self.is_qualified_name:
            spec = self._get_details_from_qualified_name()

            schema = spec['schema']
            table = spec['table']
            cluster = spec['cluster']
            column = spec['column']

            source = self._database.replace('column', 'table') if self._database else ''

            return f'{source}://{cluster}.{schema}/{table}/{column}'
        elif self.is_amundsen_key:
            return self._raw_identifier
        else:
            raise ValueError(f'Value is neither qualified name nor amundsen key: {self._raw_identifier}')



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import string
from typing import Any, List, Optional

from amundsen_common.models.table import (Application, Column,
                                          ProgrammaticDescription, Stat, Table,
                                          Tag)
from amundsen_common.models.user import User


class Fixtures:
    """
    These fixtures are useful for creating test objects. For an example usage, check out tests/tests/test_fixtures.py
    """
    counter = 1000

    @staticmethod
    def next_int() -> int:
        i = Fixtures.counter
        Fixtures.counter += 1
        return i

    @staticmethod
    def next_string(*, prefix: str = '', length: int = 10) -> str:
        astr: str = prefix + \
            ''.join(Fixtures.next_item(items=list(string.ascii_lowercase)) for _ in range(length)) + \
            ('%06d' % Fixtures.next_int())
        return astr

    @staticmethod
    def next_range() -> range:
        return range(0, Fixtures.next_int() % 5)

    @staticmethod
    def next_item(*, items: List[Any]) -> Any:
        return items[Fixtures.next_int() % len(items)]

    @staticmethod
    def next_database() -> str:
        return Fixtures.next_item(items=list(["database1", "database2"]))

    @staticmethod
    def next_application(*, application_id: Optional[str] = None) -> Application:
        if not application_id:
            application_id = Fixtures.next_string(prefix='ap', length=8)
        application = Application(application_url=f'https://{application_id}.example.com',
                                  description=f'{application_id} description',
                                  name=application_id.capitalize(),
                                  id=application_id)
        return application

    @staticmethod
    def next_tag(*, tag_name: Optional[str] = None) -> Tag:
        if not tag_name:
            tag_name = Fixtures.next_string(prefix='ta', length=8)
        return Tag(tag_name=tag_name, tag_type='default')

    @staticmethod
    def next_tags() -> List[Tag]:
        return sorted([Fixtures.next_tag() for _ in Fixtures.next_range()])

    @staticmethod
    def next_description_source() -> str:
        return Fixtures.next_string(prefix='de', length=8)

    @staticmethod
    def next_description(*, text: Optional[str] = None, source: Optional[str] = None) -> ProgrammaticDescription:
        if not text:
            text = Fixtures.next_string(length=20)
        if not source:
            source = Fixtures.next_description_source()
        return ProgrammaticDescription(text=text, source=source)

    @staticmethod
    def next_col_type() -> str:
        return Fixtures.next_item(items=['varchar', 'int', 'blob', 'timestamp', 'datetime'])

    @staticmethod
    def next_column(*,
                    table_key: str,
                    sort_order: int,
                    name: Optional[str] = None) -> Column:
        if not name:
            name = Fixtures.next_string(prefix='co', length=8)

        return Column(name=name,
                      description=f'{name} description',
                      col_type=Fixtures.next_col_type(),
                      key=f'{table_key}/{name}',
                      sort_order=sort_order,
                      stats=[Stat(stat_type='num_rows',
                                  stat_val=f'{Fixtures.next_int() * 100}',
                                  start_epoch=None,
                                  end_epoch=None)])

    @staticmethod
    def next_columns(*,
                     table_key: str,
                     randomize_pii: bool = False,
                     randomize_data_subject: bool = False) -> List[Column]:
        return [Fixtures.next_column(table_key=table_key,
                                     sort_order=i
                                     ) for i in Fixtures.next_range()]

    @staticmethod
    def next_descriptions() -> List[ProgrammaticDescription]:
        return sorted([Fixtures.next_description() for _ in Fixtures.next_range()])

    @staticmethod
    def next_table(table: Optional[str] = None,
                   cluster: Optional[str] = None,
                   schema: Optional[str] = None,
                   database: Optional[str] = None,
                   tags: Optional[List[Tag]] = None,
                   application: Optional[Application] = None) -> Table:
        """
        Returns a table for testing in the test_database
        """
        if not database:
            database = Fixtures.next_database()

        if not table:
            table = Fixtures.next_string(prefix='tb', length=8)

        if not cluster:
            cluster = Fixtures.next_string(prefix='cl', length=8)

        if not schema:
            schema = Fixtures.next_string(prefix='sc', length=8)

        if not tags:
            tags = Fixtures.next_tags()

        table_key: str = f'{database}://{cluster}.{schema}/{table}'
        # TODO: add owners, watermarks, last_udpated_timestamp, source
        return Table(database=database,
                     cluster=cluster,
                     schema=schema,
                     name=table,
                     key=table_key,
                     tags=tags,
                     table_writer=application,
                     table_readers=[],
                     description=f'{table} description',
                     programmatic_descriptions=Fixtures.next_descriptions(),
                     columns=Fixtures.next_columns(table_key=table_key),
                     is_view=False
                     )

    @staticmethod
    def next_user(*, user_id: Optional[str] = None, is_active: bool = True) -> User:
        last_name = ''.join(Fixtures.next_item(items=list(string.ascii_lowercase)) for _ in range(6)).capitalize()
        first_name = Fixtures.next_item(items=['alice', 'bob', 'carol', 'dan']).capitalize()
        if not user_id:
            user_id = Fixtures.next_string(prefix='us', length=8)
        return User(user_id=user_id,
                    email=f'{user_id}@example.com',
                    is_active=is_active,
                    first_name=first_name,
                    last_name=last_name,
                    full_name=f'{first_name} {last_name}')


def next_application(**kwargs: Any) -> Application:
    return Fixtures.next_application(**kwargs)


def next_int() -> int:
    return Fixtures.next_int()


def next_string(**kwargs: Any) -> str:
    return Fixtures.next_string(**kwargs)


def next_range() -> range:
    return Fixtures.next_range()


def next_item(**kwargs: Any) -> Any:
    return Fixtures.next_item(**kwargs)


def next_database() -> str:
    return Fixtures.next_database()


def next_tag(**kwargs: Any) -> Tag:
    return Fixtures.next_tag(**kwargs)


def next_tags() -> List[Tag]:
    return Fixtures.next_tags()


def next_description_source() -> str:
    return Fixtures.next_description_source()


def next_description(**kwargs: Any) -> ProgrammaticDescription:
    return Fixtures.next_description(**kwargs)


def next_col_type() -> str:
    return Fixtures.next_col_type()


def next_column(**kwargs: Any) -> Column:
    return Fixtures.next_column(**kwargs)


def next_columns(**kwargs: Any) -> List[Column]:
    return Fixtures.next_columns(**kwargs)


def next_descriptions() -> List[ProgrammaticDescription]:
    return Fixtures.next_descriptions()


def next_table(**kwargs: Any) -> Table:
    return Fixtures.next_table(**kwargs)


def next_user(**kwargs: Any) -> User:
    return Fixtures.next_user(**kwargs)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import getpass

from flask import current_app as flask_app

from amundsen_common.log.caller_retrieval import BaseCallerRetriever


class AuthCallerRetrieval(BaseCallerRetriever):
    def get_caller(self) -> str:
        if flask_app.config.get('AUTH_USER_METHOD', None):
            return flask_app.config['AUTH_USER_METHOD'](flask_app).email
        return getpass.getuser()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from enum import Enum, auto


class ResourceType(Enum):
    Table = auto()
    Dashboard = auto()
    User = auto()
    Column = auto()
    Type_Metadata = auto()
    Feature = auto()


def to_resource_type(*, label: str) -> ResourceType:
    return ResourceType[label.title()]


def to_label(*, resource_type: ResourceType) -> str:
    return resource_type.name.lower()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

"""
An Action Logger module. Singleton pattern has been applied into this module
so that registered callbacks can be used all through the same python process.
"""

import logging
import sys
from typing import Callable, List, Any

from pkg_resources import iter_entry_points

from amundsen_common.log.action_log_model import ActionLogParams

LOGGER = logging.getLogger(__name__)

__pre_exec_callbacks = []  # type: List[Callable[..., Any]]
__post_exec_callbacks = []  # type: List[Callable[..., Any]]


def register_pre_exec_callback(action_log_callback: Callable[..., Any]) -> None:
    """
    Registers more action_logger function callback for pre-execution. This function callback is expected to be called
    with keyword args. For more about the arguments that is being passed to the callback, refer to
    amundsen_application.log.action_log_model.ActionLogParams
    :param action_logger: An action logger callback function
    :return: None
    """
    LOGGER.debug("Adding {} to pre execution callback".format(action_log_callback))
    __pre_exec_callbacks.append(action_log_callback)


def register_post_exec_callback(action_log_callback: Callable[..., Any]) -> None:
    """
    Registers more action_logger function callback for post-execution. This function callback is expected to be
    called with keyword args. For more about the arguments that is being passed to the callback,
    amundsen_application.log.action_log_model.ActionLogParams
    :param action_logger: An action logger callback function
    :return: None
    """
    LOGGER.debug("Adding {} to post execution callback".format(action_log_callback))
    __post_exec_callbacks.append(action_log_callback)


def on_pre_execution(action_log_params: ActionLogParams) -> None:
    """
    Calls callbacks before execution.
    Note that any exception from callback will be logged but won't be propagated.
    :param kwargs:
    :return: None
    """
    LOGGER.debug("Calling callbacks: {}".format(__pre_exec_callbacks))
    for call_back_function in __pre_exec_callbacks:
        try:
            call_back_function(action_log_params)
        except Exception:
            logging.exception('Failed on pre-execution callback using {}'.format(call_back_function))


def on_post_execution(action_log_params: ActionLogParams) -> None:
    """
    Calls callbacks after execution. As it's being called after execution, it can capture most of fields in
    amundsen_application.log.action_log_model.ActionLogParams. Note that any exception from callback will be logged
    but won't be propagated.
    :param kwargs:
    :return: None
    """
    LOGGER.debug("Calling callbacks: {}".format(__post_exec_callbacks))
    for call_back_function in __post_exec_callbacks:
        try:
            call_back_function(action_log_params)
        except Exception:
            logging.exception('Failed on post-execution callback using {}'.format(call_back_function))


def logging_action_log(action_log_params: ActionLogParams) -> None:
    """
    An action logger callback that just logs the ActionLogParams that it receives.
    :param **kwargs keyword arguments
    :return: None
    """
    if LOGGER.isEnabledFor(logging.DEBUG):
        LOGGER.debug('logging_action_log: {}'.format(action_log_params))


def register_action_logs() -> None:
    """
    Retrieve declared action log callbacks from entry point where there are two groups that can be registered:
     1. "action_log.post_exec.plugin": callback for pre-execution
     2. "action_log.pre_exec.plugin": callback for post-execution
    :return: None
    """
    for entry_point in iter_entry_points(group='action_log.post_exec.plugin', name=None):
        print('Registering post_exec action_log entry_point: {}'.format(entry_point), file=sys.stderr)
        register_post_exec_callback(entry_point.load())

    for entry_point in iter_entry_points(group='action_log.pre_exec.plugin', name=None):
        print('Registering pre_exec action_log entry_point: {}'.format(entry_point), file=sys.stderr)
        register_pre_exec_callback(entry_point.load())


register_action_logs()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional, Dict, Any

import attr

from amundsen_common.models.user import User
from amundsen_common.models.badge import Badge
from amundsen_common.models.tag import Tag
from amundsen_common.models.table import ProgrammaticDescription
from marshmallow3_annotations.ext.attrs import AttrsSchema


@attr.s(auto_attribs=True, kw_only=True)
class ColumnItem:
    column_name: str
    column_type: str


class ColumnItemSchema(AttrsSchema):
    class Meta:
        target = ColumnItem
        register_as_scheme = True


@attr.s(auto_attribs=True, kw_only=True)
class DataSample:
    # Modeled after preview data model in FE
    columns: List[ColumnItem]
    data: List[Dict[str, Any]]
    error_text: str


class DataSampleSchema(AttrsSchema):
    class Meta:
        target = DataSample
        register_as_scheme = True


@attr.s(auto_attribs=True, kw_only=True)
class FeatureWatermark:
    key: Optional[str]
    watermark_type: Optional[str]
    time: str


class FeatureWatermarkSchema(AttrsSchema):
    class Meta:
        target = FeatureWatermark
        register_as_scheme = True


@attr.s(auto_attribs=True, kw_only=True)
class Feature:
    key: Optional[str] = attr.ib(default=None)
    name: str
    version: str  # ex: "1.2.0"
    status: Optional[str]
    feature_group: str
    entity: Optional[str]
    data_type: Optional[str]
    availability: List[str]
    description: Optional[str] = attr.ib(default=None)
    owners: List[User]
    badges: List[Badge]
    tags: List[Tag]
    programmatic_descriptions: List[ProgrammaticDescription]
    watermarks: List[FeatureWatermark]
    last_updated_timestamp: Optional[int]
    created_timestamp: Optional[int]


class FeatureSchema(AttrsSchema):
    class Meta:
        target = Feature
        register_as_scheme = True


@attr.s(auto_attribs=True, kw_only=True)
class FeatureSummary:
    key: str  # ex: test_feature_group_name/test_feature_name/1.2.0
    name: str
    version: str
    availability: List[str]
    entity: List[str]
    description: Optional[str] = attr.ib(default=None)
    badges: List[Badge]
    last_updated_timestamp: Optional[int]


class FeatureSummarySchema(AttrsSchema):
    class Meta:
        target = FeatureSummary
        register_as_scheme = True
